function ComputeFlow
% flow computation with ABD
% If you use this code, please cite the following paper:
% N. Ray, Computation of fluid and particle motion from time sequenced
% image pair: a global outlier identification approach, IEEE Transactions 
% on Image Processing, vol.20, no.10, pp.2925-2936, 2011.

% Author: Nilanjan Ray
% Use the code anyway you want!

close all;
clear all;

% read image pair
 
I = double(imread('piv01_1.bmp'));
J = double(imread('piv01_2.bmp'));

figure(1);imagesc(I),colormap(gray),axis image;
figure(2);imagesc(J),colormap(gray),axis image;

% parameters - as per the above paper
lambda=0.15;
alpha=5*lambda;
beta=2*alpha;
level=6;
factor=1.5;
frac = 0.25; % referred to as 1-(K/M) as per the above paper.

tic;
[Fx,Fy]=OpticalFlow(I,J,lambda,alpha,beta,level,factor,frac);
toc;

[h,w] = size(I);
[X,Y]=meshgrid(1:w,1:h);
figure(3), quiver(X(1:8:end,1:8:end),Y(1:8:end,1:8:end),Fx(1:8:end,1:8:end),Fy(1:8:end,1:8:end),'r'), axis ij, axis image;

function [Fx,Fy]=OpticalFlow(I,J,lambda,alpha,beta,level,factor,frac)
% I: image 1
% J: image 2
% alpha: smoothing parameter
% level: highest level of pyramid
%[Fx,Fy]: computed optical flow

% makes images double
I = double(I);
J = double(J);

% normalize intensity range
I(:) = (I-min(I(:)))/(max(I(:))-min(I(:)));
J(:) = (J-min(J(:)))/(max(J(:))-min(J(:)));

% compute Gaussian pyramid
GaussianPyramidI = GaussianPyramid(I,level,factor);
GaussianPyramidJ = GaussianPyramid(J,level,factor);

% multiresolution processing
for n=level:-1:1
    disp(['Processing at pyramid level: ' num2str(n)]);
    [h,w] = size(GaussianPyramidJ{n});
    if n==level,
        % compute optical flow at the highest of the pyramid
         [Fx,Fy] = BCDFlow(GaussianPyramidI{n},GaussianPyramidJ{n},lambda,alpha,beta,level,frac);
    else
        % upsample flow
        Fx = imresize(factor*Fx,[h w],'bilinear','Antialiasing',true);
        Fy = imresize(factor*Fy,[h w],'bilinear','Antialiasing',true);
        % warp second image
        [X,Y] = meshgrid(1:w,1:h);
        thisJ =  interp2(X,Y,GaussianPyramidJ{n},X+Fx,Y+Fy,'*linear',0);
        clear X; clear Y;
        % compute flow correction
         [U,V] = BCDFlow(GaussianPyramidI{n},thisJ,lambda,alpha,beta,n,frac);
        clear thisJ;
        % add correction
        Fx(:) = Fx + U;
        Fy(:) = Fy + V;
        clear U; clear V;
    end
    disp(' ');
end

function GI = GaussianPyramid(I,level,factor)
% return a cell array
GI = cell(level,1);
GI{1} = imfilter(I,fspecial('Gaussian',[7 7],1),'same','conv','replicate');
for n=2:level,
    h = round(size(GI{n-1},1)/factor);
    w = round(size(GI{n-1},2)/factor);
    GI{n} = imresize(GI{n-1},[h w],'bicubic','Antialiasing',true);
end

function [Fx,Fy] = BCDFlow(I,J,lambda,alpha,beta,level,frac)

[h,w] = size(I);

% for photometric matching consider a block size of (2*hf+1) x (2*wf+1)
hf = 2;    % window size for matching - this is a tunable parameter
wf = 2;    % window size for matching - this is a tunable parameter

% maximum displacement
if level>1,
    threshold = 1.5;  % this is a tunable parameter
    stepsize=0.25;
else
    threshold = 0.75;  % this is a tunable parameter
    stepsize=0.125;
    hf = 1;
    wf = 1;
end
% threshold = 1.5;
% stepsize = 0.25;
if level==-1,
    threshold = 0.375;
    stepsize = 0.0625;
    hf = 1;    % window size for matching - this is a tunable parameter
    wf = 1;    % window size for matching - this is a tunable parameter
end
if level==-2,
    threshold = 0.1875;
    stepsize = 0.0313;
    hf = 1;    % window size for matching - this is a tunable parameter
    wf = 1;    % window size for matching - this is a tunable parameter
end

if level==-3,
    threshold = 0.01;
    stepsize = 0.002;
    hf = 1;    % window size for matching - this is a tunable parameter
    wf = 1;    % window size for matching - this is a tunable parameter
end

[X,Y] = meshgrid(-threshold:stepsize:threshold,-threshold:stepsize:threshold);
X = X(:); Y = Y(:);

% Photometric part of the cost function
Pc = 1e10*ones(h*w,length(X));

% Cost matrix computation - implements sum of absolute value 
% difference in a window size of (2*hf+1)x(2*wf+1)
[W,H] = meshgrid(1:w,1:h);
T = zeros(h,w);
% for NCC
TI=T;
TJ=T;
TIJ=T;
for n=1:length(X),
    T(:) = interp2(W,H,J,W+X(n),H+Y(n),'*linear',-255);
    TIJ(:) = imfilter((I.*T),fspecial('average',[2*hf+1 2*wf+1]),'same','conv','replicate');
    TI(:)  = imfilter((I.*I),fspecial('average',[2*hf+1 2*wf+1]),'same','conv','replicate');
    TJ(:)  = imfilter((T.*T),fspecial('average',[2*hf+1 2*wf+1]),'same','conv','replicate');
    T(:)   = TIJ./(sqrt(TI).*sqrt(TJ)+eps);
    T(:)   = (1-T)/2;
    Pc(:,n) = T(:);
end

% create a mask to indicate the borders of the displacement flow field
Mask = zeros(h,w); Mask(1,:)=1; Mask(end,:)=1; Mask(:,1) = 1; Mask(:,end) = 1; 
[Fx,Fy,Gx,Gy] = DivCurl_OF_solve(Pc,X,Y,logical(Mask),lambda,alpha,beta,h,w,frac); % for outliers
Fx = reshape(Fx,h,w);
Fy = reshape(Fy,h,w);

function [Fx,Fy,Gx,Gy] = DivCurl_OF_solve(Pc,X,Y,Mask,lambda,alpha,beta,r1,c1,frac)
% solves the energy functional with assignment variables and displacement
% parameters
% The assignment problem here is solved by maximum bipartite matching
% Pc: Photometric cost, such as Pc(i,j) stores the ssd betwwen image patches at i and j
% Mask: logical matrix to denote the borders of the displacement field Fx and Fy
% X,Y:  stores offset information
% r1,c1: size of velocity vector field
% frac: outlier parameter

[n,m] = size(Pc);

% Fx and Fy are displacement fields we want to compute
Fx = zeros(n,1);
Fy = zeros(n,1);
Gx = zeros(n,1);
Gy = zeros(n,1);
Gxx = zeros(r1,c1);
Gxy = Gxx;
Gyx = Gxx;
Gyy = Gxx;
% Compute the initial assignment based on just the photometric cost Pc

[minPc,a] = min(Pc,[],2);

% Block coordinate descent (i.e., alternating optimization)
Cost = zeros(n,m);
bx = zeros(n,1);
by = zeros(n,1);

Dx = repmat(X',n,1);
Dy = repmat(Y',n,1);

d = zeros(n,1);

for itr=1:10,
    
    K = round(frac*n);
    
    d(:) = (lambda/alpha);
    % Attempting to control discontinuities/outliers
    [sortedCost,sortedInd] = sort(minPc,'descend');
    % clear K worst costs
    if K>0,
        d(sortedInd(1:K))=0;
    end
    % solve for Fx and Fy
    bx(:) = d.*Dx((1:n)'+(a-1)*n) + (-Gxx(:)+Gyy(:));
    by(:) = d.*Dy((1:n)'+(a-1)*n) + (-Gxy(:)-Gyx(:));
    
    Fx(:) = pde_solve(bx,d,Mask); % Uses Neumann BC
    Fy(:) = pde_solve(by,d,Mask); % Uses Neumann BC
    
    [Fxx,Fxy] = gradient(reshape(Fx,r1,c1));
    [Fyx,Fyy] = gradient(reshape(Fy,r1,c1));
    
    % solve for Gx, Gy
    bx(:) = (alpha/beta)*(Fxx(:)+Fyy(:));
    by(:) = (alpha/beta)*(Fyx(:)-Fxy(:));
    d(:)  = (alpha/beta)*ones(n,1);
    
    Gx(:) = pde_solve(bx,d,Mask); % Uses Neumann BC
    Gy(:) = pde_solve(by,d,Mask); % Uses Neumann BC
    
    [Gxx,Gxy] = gradient(reshape(Gx,r1,c1));
    [Gyx,Gyy] = gradient(reshape(Gy,r1,c1));
    
    % solve for assignment a
    Cost(:) = Pc + 0.5*lambda*((Dx-repmat(Fx,1,m)).^2+(Dy-repmat(Fy,1,m)).^2);
    
    clear a; clear minPc;
    [minPc,a] = min(Cost,[],2);
    disp(['itr ' num2str(itr)]);
end

function x=pde_solve(b,d,B)
% solves a pde: laplacian(x) + (d)x = b, subject to Neumann BCs 
% input parameters: b is the right hand side of pde above, also it appears
% on the rhs of Ax = b equation, it is a vector
% d: is the diagonal to be added to the diagonal of the Laplacian matrix,
% it is a vector; d appears in the above pde as shown
% B is the mask for border pixels, it is a 0-1 (logical) matrix
% x is the output, it is a vector

[h, w]=size(B);

% logical indices for interior points
M = logical(~B);

D5 = zeros(h*w,5);
% Laplacian for interior points
D5(M(:),1)=-1; D5(M(:),2)=-1;  D5(M(:),3)=4+d(M(:)); D5(M(:),4)=-1; D5(M(:),5)=-1;

% Neumann BC for border points
D5(B(:),3)=3;
% assign 2 for the four corner points
D5(1,3)=2; D5(h,3) = 2; D5(h*(w-1)+1,3)=2; D5(h*w,3)=2;

lB = zeros(h,w); lB(:,1) = 1; lB = logical(lB); % left border
rB = zeros(h,w); rB(:,end) = 1; rB = logical(rB); % right border
uB = zeros(h,w); uB(1,2:w-1) = 1; uB = logical(uB); % top border
dB = zeros(h,w); dB(end,2:w-1) = 1; dB = logical(dB); % bottom border

D5(lB(:),1)=-1; D5(lB(:),2)=-1; D5(lB(:),4)=-1;
D5(rB(:),5)=-1; D5(rB(:),2)=-1; D5(rB(:),4)=-1;
D5(uB(:),2)=-1; D5(uB(:),1)=-1; D5(uB(:),5)=-1;
D5(dB(:),4)=-1; D5(dB(:),1)=-1; D5(dB(:),5)=-1;

% take care of two corners
D5(h,2)=0; D5(h*(w-1)+1,4)=0;

b(B(:))=0; % right hand side border pixels cleared for Neumann BC

% sparse matrix
A = spdiags(D5,[-h -1 0 1 h],h*w,h*w)';

% Solve for x
x = A\b;